@@ -1,5 +1,5 @@
 package PAR;
-$PAR::VERSION = '0.980';
+$PAR::VERSION = '0.983';
 
 use 5.006;
 use strict;
@@ -26,13 +26,16 @@ BEGIN {
     }
 }
 
+use PAR::SetupProgname;
+use PAR::SetupTemp;
+
 =head1 NAME
 
 PAR - Perl Archive Toolkit
 
 =head1 VERSION
 
-This document describes version 0.980 of PAR, released May 22, 2008.
+This document describes release 0.983 of PAR, released September 12, 2008.
 
 =head1 SYNOPSIS
 
@@ -314,7 +317,6 @@ my $is_insensitive_fs = (
         and (-s lc($progname) || -1) == (-s uc($progname) || -1)
         and (-s lc($progname) || -1) == -s $progname
 );
-my $par_temp;
 
 # lexical for import(), and _import_foo() functions to control unpar()
 my %unpar_options;
@@ -323,8 +325,8 @@ my %unpar_options;
 sub import {
     my $class = shift;
 
-    _set_progname();
-    _set_par_temp();
+    PAR::SetupProgname::set_progname();
+    PAR::SetupTemp::set_par_temp_env();
 
     $progname = $ENV{PAR_PROGNAME} ||= $0;
     $is_insensitive_fs = (-s $progname and (-s lc($progname) || -1) == (-s uc($progname) || -1));
@@ -485,7 +487,7 @@ sub _import_hash_ref {
             croak("Cannot run script '$script' from PAR file '$opt->{file}'. Script couldn't be found in PAR file.");
         }
         
-        _run_member($member);
+        _run_member_from_par($member);
     }
 
     return();
@@ -534,6 +536,27 @@ sub _first_member {
     return;
 }
 
+sub _run_member_from_par {
+    my $member = shift;
+    my $clear_stack = shift;
+    my ($fh, $is_new, $filename) = _tempfile($member->crc32String . ".pl");
+
+    if ($is_new) {
+        my $file = $member->fileName;
+        print $fh "package main;\n";
+        print $fh "#line 1 \"$file\"\n";
+        $member->extractToFileHandle($fh);
+        seek ($fh, 0, 0);
+    }
+
+    $ENV{PAR_0} = $filename; # for Pod::Usage
+    { do $filename;
+      CORE::exit($1) if ($@ =~/^_TK_EXIT_\((\d+)\)/);
+      die $@ if $@;
+      exit;
+    }
+}
+
 sub _run_member {
     my $member = shift;
     my $clear_stack = shift;
@@ -560,39 +583,63 @@ sub _run_member {
     }
 }
 
+# extract the contents of a .par (or .exe) or any
+# Archive::Zip handle to the PAR_TEMP/inc directory.
+# returns that directory.
 sub _extract_inc {
-    my $file = shift;
-    my $inc = "$par_temp/inc";
+    my $file_or_azip_handle = shift;
+    my $force_extract = shift;
+    my $inc = "$PAR::SetupTemp::PARTemp/inc";
     my $dlext = defined($Config{dlext}) ? $Config::Config{dlext} : '';
+    my $inc_exists = -d $inc;
+    my $is_handle = ref($file_or_azip_handle) && $file_or_azip_handle->isa('Archive::Zip::Archive');
 
-    if (!-d $inc) {
+    require File::Spec;
+
+    if (!$inc_exists or $force_extract) {
         for (1 .. 10) { mkdir("$inc.lock", 0755) and last; sleep 1 }
         
-        # First try to unzip the *fast* way.
-        eval {
-          require Archive::Unzip::Burst;
-          Archive::Unzip::Burst::unzip($file, $inc)
-            and die "Could not unzip '$file' into '$inc'. Error: $!";
-        };
+        undef $@;
+        if (!$is_handle) {
+          # First try to unzip the *fast* way.
+          eval {
+            require Archive::Unzip::Burst;
+            Archive::Unzip::Burst::unzip($file_or_azip_handle, $inc)
+              and die "Could not unzip '$file_or_azip_handle' into '$inc'. Error: $!";
+              die;
+          };
 
-        # This means the fast module is there, but didn't work.
-        if ($@ =~ /^Could not unzip/) {
-          die $@;
+          # This means the fast module is there, but didn't work.
+          if ($@ =~ /^Could not unzip/) {
+            die $@;
+          }
         }
 
-        # failed to load Archive::Unzip::Burst. Default to slow way.
-        elsif ($@) {
-          open my $fh, '<', $file or die "Cannot find '$file': $!";
-          binmode($fh);
-          bless($fh, 'IO::File');
+        # either failed to load Archive::Unzip::Burst or got an A::Zip handle
+        # fallback to slow way.
+        if ($is_handle || $@) {
+          my $zip;
+          if (!$is_handle) {
+            open my $fh, '<', $file_or_azip_handle
+              or die "Cannot find '$file_or_azip_handle': $!";
+            binmode($fh);
+            bless($fh, 'IO::File');
 
-          my $zip = Archive::Zip->new;
-          ( $zip->readFromFileHandle($fh, $file) == Archive::Zip::AZ_OK() )
-              or die "Read '$file' error: $!";
+            $zip = Archive::Zip->new;
+            ( $zip->readFromFileHandle($fh, $file_or_azip_handle) == Archive::Zip::AZ_OK() )
+                or die "Read '$file_or_azip_handle' error: $!";
+          }
+          else {
+            $zip = $file_or_azip_handle;
+          }
+
+          mkdir($inc) if not -d $inc;
 
           for ( $zip->memberNames() ) {
               next if m{\.\Q$dlext\E[^/]*$};
               s{^/}{};
+              my $outfile =  File::Spec->catfile($inc, $_);
+              next if -e $outfile and not -w _;
               $zip->extractMember($_, "$inc/" . $_);
           }
         }
@@ -600,9 +647,16 @@ sub _extract_inc {
         rmdir("$inc.lock");
     }
 
-    require File::Spec;
-    unshift @INC, grep -d, map File::Spec->catdir($inc, @$_),
-        [ 'lib' ], [ 'arch' ], [ $arch ], [ $ver ], [ $ver, $arch ], [];
+    # add the freshly extracted directories to @INC,
+    # but make sure there's no duplicates
+    my %inc_exists = map { ($_, 1) } @INC;
+    unshift @INC, grep !exists($inc_exists{$_}),
+                  grep -d,
+                  map File::Spec->catdir($inc, @$_),
+                  [ 'lib' ], [ 'arch' ], [ $arch ],
+                  [ $ver ], [ $ver, $arch ], [];
+
+    return $inc;
 }
 
 
@@ -815,12 +869,12 @@ sub unpar {
         # hook, so no need to pull it out here.
         # Allow this to be disabled so caller can do their own caching
         # via import({no_shlib_unpack => 1, file => foo.par})
-        unless($unpar_options{no_shlib_unpack}) {
+        if(not $unpar_options{no_shlib_unpack} and defined $ENV{PAR_TEMP}) {
             my @members = _cached_members_matching( $zip,
               qr#^shlib/$Config{archname}/.*\.\Q$Config{dlext}\E(?:\.|$)#
             );
             foreach my $member (@members) {
-                next if $member->isDirectory or !$ENV{PAR_TEMP};
+                next if $member->isDirectory;
                 my $member_name = $member->fileName;
                 next unless $member_name =~ m{
                         \/([^/]+)$
@@ -888,75 +942,6 @@ sub unpar {
     return $fh;
 }
 
-# The C version of this code appears in myldr/mktmpdir.c
-sub _set_par_temp {
-    if ($ENV{PAR_TEMP} and $ENV{PAR_TEMP} =~ /(.+)/) {
-        $par_temp = $1;
-        return;
-    }
-
-    require File::Spec;
-
-    foreach my $path (
-        (map $ENV{$_}, qw( PAR_TMPDIR TMPDIR TEMPDIR TEMP TMP )),
-        qw( C:\\TEMP /tmp . )
-    ) {
-        next unless $path and -d $path and -w $path;
-        my $username;
-        my $pwuid;
-        # does not work everywhere:
-        eval {($pwuid) = getpwuid($>) if defined $>;};
-
-        if ( defined(&Win32::LoginName) ) {
-            $username = &Win32::LoginName;
-        }
-        elsif (defined $pwuid) {
-            $username = $pwuid;
-        }
-        else {
-            $username = $ENV{USERNAME} || $ENV{USER} || 'SYSTEM';
-        }
-        $username =~ s/\W/_/g;
-
-        my $stmpdir = File::Spec->catdir($path, "par-$username");
-        ($stmpdir) = $stmpdir =~ /^(.*)$/s;
-        mkdir $stmpdir, 0755;
-        if (!$ENV{PAR_CLEAN} and my $mtime = (stat($progname))[9]) {
-            my $ctx = eval { require Digest::SHA; Digest::SHA->new(1) }
-                   || eval { require Digest::SHA1; Digest::SHA1->new }
-                   || eval { require Digest::MD5; Digest::MD5->new };
-
-            # Workaround for bug in Digest::SHA 5.38 and 5.39
-            my $sha_version = eval { $Digest::SHA::VERSION } || 0;
-            if ($sha_version eq '5.38' or $sha_version eq '5.39') {
-                $ctx->addfile($progname, "b") if ($ctx);
-            }
-            else {
-                if ($ctx and open(my $fh, "<$progname")) {
-                    binmode($fh);
-                    $ctx->addfile($fh);
-                    close($fh);
-                }
-            }
-
-            $stmpdir = File::Spec->catdir(
-                $stmpdir,
-                "cache-" . ( $ctx ? $ctx->hexdigest : $mtime )
-            );
-        }
-        else {
-            $ENV{PAR_CLEAN} = 1;
-            $stmpdir = File::Spec->catdir($stmpdir, "temp-$$");
-        }
-
-        $ENV{PAR_TEMP} = $stmpdir;
-        mkdir $stmpdir, 0755;
-        last;
-    }
-
-    $par_temp = $1 if $ENV{PAR_TEMP} and $ENV{PAR_TEMP} =~ /(.+)/;
-}
-
 sub _tempfile {
     my ($fh, $filename);
     if ($ENV{PAR_CLEAN} or !@_) {
@@ -967,7 +952,7 @@ sub _tempfile {
             # and will be deleted by the C runtime; having File::Temp
             # delete it has the only effect of giving ugly warnings
             ($fh, $filename) = File::Temp::tempfile(
-                DIR     => $par_temp,
+                DIR     => $PAR::SetupTemp::PARTemp,
                 UNLINK  => ($^O ne 'MSWin32' and $^O !~ /hpux/),
             ) or die "Cannot create temporary file: $!";
             binmode($fh);
@@ -978,7 +963,7 @@ sub _tempfile {
     require File::Spec;
 
     # untainting tempfile path
-    local $_ = File::Spec->catfile( $par_temp, $_[0] );
+    local $_ = File::Spec->catfile( $PAR::SetupTemp::PARTemp, $_[0] );
     /^(.+)$/ and $filename = $1;
 
     if (-r $filename) {
@@ -992,34 +977,6 @@ sub _tempfile {
     return ($fh, 1, $filename);
 }
 
-sub _set_progname {
-    require File::Spec;
-
-    if ($ENV{PAR_PROGNAME} and $ENV{PAR_PROGNAME} =~ /(.+)/) {
-        $progname = $1;
-    }
-    $progname ||= $0;
-
-    if (( () = File::Spec->splitdir($progname) ) > 1 or !$ENV{PAR_PROGNAME}) {
-        if (open my $fh, $progname) {
-            return if -s $fh;
-        }
-        if (-s "$progname$Config{_exe}") {
-            $progname .= $Config{_exe};
-            return;
-        }
-    }
-
-    foreach my $dir (split /\Q$Config{path_sep}\E/, $ENV{PATH}) {
-        next if exists $ENV{PAR_TEMP} and $dir eq $ENV{PAR_TEMP};
-        my $name = File::Spec->catfile($dir, "$progname$Config{_exe}");
-        if (-s $name) { $progname = $name; last }
-        $name = File::Spec->catfile($dir, "$progname");
-        if (-s $name) { $progname = $name; last }
-    }
-}
-
-
 # Given an Archive::Zip object, this generates a hash of
 #   file_name_in_zip => file object
 # and returns a reference to that.
